#include <allegro.h>
#include <cstring>
#include <cmath>
#include <iostream>
#include <fstream>
#include <list>

using namespace std;

int logic_time = 0;
int anim_time = 0;

void logic_timer()
{
  logic_time++;
}
END_OF_FUNCTION(logic_timer);

void anim_timer()
{
  anim_time++;
}
END_OF_FUNCTION(anim_timer);

#define HEIGHT (3*SCREEN_H/5)


bool quitnow = false;
void Close() {
  quitnow = true;
}

// initialize systems!
BITMAP* init(int w, int h) {
  
  allegro_init();
  install_mouse();
  install_timer();
  install_keyboard();
  
  set_color_depth(16);
  if (set_gfx_mode(GFX_AUTODETECT_WINDOWED, w, h, 0, 0) != 0) {
    set_color_depth(15);
    if (set_gfx_mode(GFX_AUTODETECT_WINDOWED, w, h, 0, 0) != 0){
      allegro_message("Couldn't set graphics mode!");
      exit(1);
    }
  }
  
  char buf[1024];
  get_executable_name(buf, 1010);
  char *fname = (get_filename(buf));
  strcpy(fname, "/data/"); // buffer overflow?  What's that?
  chdir(buf);
  
  LOCK_VARIABLE(logic_time);
  LOCK_VARIABLE(anim_time);
  LOCK_FUNCTION(logic_timer);
  LOCK_FUNCTION(anim_timer);
  install_int_ex(logic_timer, BPS_TO_TIMER(60));
  
  set_window_close_hook(Close);
  set_window_title("Ninja Ant: Survive!");
  
  return create_bitmap(w, h);
}

struct Level {
  int *heights;
  
  Level() {
    heights = new int[SCREEN_W];
    if (!heights) {
      cout << "OH NO!  Could not create our height field."  << endl;
      exit(1);
    }
    for (int i = 0; i < SCREEN_W; i++) {
      heights[i] = 0;
    }
  }
  ~Level() {
    delete heights;
  }
  
  int MaxY(int minX, int maxX) {
    if (maxX >= SCREEN_W) maxX = SCREEN_W-1;
    if (minX < 0) minX = 0;
      
    if (maxX <= minX) {
      cout << "OH NO!  Invalid input to MaxY caused me to die." << endl;
      exit(1);
    }
    
    int maxY = heights[minX];
    for (int x = minX+1; x <= maxX; x++) {
      if (heights[x] > maxY) {
        maxY = heights[x];
      }
    }
    
    return maxY;
  }
  
  void Crush(int minX, int maxX, int distance) {
    if (minX < 0)
      minX = 0;
    if (maxX >= SCREEN_W)
      maxX = SCREEN_W - 1;
    int newValue = MaxY(minX, maxX) - distance;
    
    for (int x = minX; x <= maxX; x++) {
      if (heights[x] > newValue)
        heights[x] = newValue;
    }
  }
  
  void Draw(BITMAP *buffer, float y) {
    for (int x = 0; x < SCREEN_W; x++) {
      vline(buffer, x, -heights[x] + int(y) + HEIGHT + 12, SCREEN_H, makecol(128,64,0));
      vline(buffer, x, -heights[x] + int(y) + HEIGHT - 12, -heights[x] + int(y) + HEIGHT + 12, makecol(255,128,0));
    }
  }
};

enum STATES { JUMP, CROUCH, STAND, RUN };


BITMAP *LoadBitmap(const char * file, RGB *palette = NULL) {
  BITMAP *tr = load_bitmap(file, palette);
  if (!tr) {
    cout << "Couldn't load bitmap: " << file << endl;
    exit(1);
  }
  return tr;
}

struct Particle {
  float x, y, v_x, v_y, gravity, jitter;
  int color, life;
  
  Particle(float x, float y, int color, float g = -.2f, float jitter = 0, int life = 40) : x(x), y(y), v_x(0), v_y(0), gravity(g), jitter(jitter), color(color), life(life) {
    while (fabs(v_x) + fabs(v_y) == 0) {
      v_x = ((rand()%7) - 3)*.5;
      v_y = ((rand()%7) - 1)*.5;
    }
    life -= rand()%3+1;
  }
  
  void Update() {
    if (life<=0)
      return;
    
    v_y += gravity;
    x += v_x;
    y += v_y;
    v_x += jitter * ((rand()%7) - 3);
    life--;
  }
  
  void Draw(BITMAP *buffer, int why) {
    if (life>0)
      putpixel(buffer, int(x), why - int(y) + HEIGHT, color);
  }
};

struct Character {
  BITMAP *jump, *crouch, *stand, *dodge;
  list<BITMAP*> run;
  
  int state, health;
  bool hit;
  int last;
  float x, y, v_x, v_y;
  int last_left, last_right;
  
  Character() {
    x = SCREEN_W/2;
    y = v_x = v_y = 0;
    state = STAND;
    health = 10;
    jump = LoadBitmap("jump.bmp", NULL);
    crouch = LoadBitmap("crouch.bmp", NULL);
    stand = LoadBitmap("still.bmp", NULL);
    dodge = LoadBitmap("dodge.bmp");
    run.push_back(LoadBitmap("run1.bmp", NULL));
    run.push_back(LoadBitmap("run2.bmp", NULL));
    run.push_back(LoadBitmap("run3.bmp", NULL));
    run.push_back(LoadBitmap("run4.bmp", NULL));
    last_left = last_right = -1000;
  }
  ~Character() {
    destroy_bitmap(jump);
    destroy_bitmap(crouch);
    destroy_bitmap(stand);
    for (list<BITMAP*>::iterator i = run.begin(); i != run.end(); ++i) {
      destroy_bitmap(*i);
    }
  }
  
  void Boom(list<Particle> &parts, int color) {
    BITMAP *reference = create_bitmap(stand->w, stand->h);
    if (!reference) {
      cout << "Couldn't create reference bitmap!" << endl;
      exit(1);
    }
    
    clear_to_color(reference, makecol(255,255,255));
    Draw(reference, false);
    
    for (int i = 0; i < reference->w; i++) {
      for (int j = 0; j < reference->h; j++) {
        if (getpixel(reference, i, reference->h-j-1) == makecol(0,0,0)) {
          parts.push_back(Particle(x+i, y+j, color));
        }
      }
    }
    
    destroy_bitmap(reference);
  }
  
  void Draw(BITMAP *buffer, bool large = true) {
    list<BITMAP*>::iterator r;
    int c = ( (anim_time/4) % 4 );
    
    int s_x = int(x) * large;
    int s_y = (HEIGHT - stand->h) * large;
    
    int dir = 0;
    if (v_x > 0)
      dir = 1;
    else if (v_x < 0)
      dir = -1;
    else
      dir = last;
    
    if (fabs(v_x) > 5.1f) {
      if (dir > 0)
        draw_sprite(buffer, dodge, s_x, s_y);
      else
        draw_sprite_h_flip(buffer, dodge, s_x, s_y);
    }
    else {
      switch (state) {
        case JUMP:
          if (dir > 0)
            rotate_sprite(buffer, jump, s_x, s_y, itofix(anim_time*30));
          else
            rotate_sprite_v_flip(buffer, jump, s_x, s_y, itofix(-anim_time*30 + 128)); 
          break;
        case RUN:
          r = run.begin();
          while (c > 0) {
            c--;
            ++r;
          }
          if (dir > 0)
            draw_sprite(buffer, *r, s_x, s_y);
          else
            draw_sprite_h_flip(buffer, *r, s_x, s_y);
          break;
        case CROUCH:
          if (dir <= 0)
            draw_sprite_h_flip(buffer, crouch, s_x, s_y);
          else
            draw_sprite(buffer, crouch, s_x, s_y);
            
          break;
        default:
          if (dir > 0)
            draw_sprite(buffer, stand, s_x, s_y);
          else
            draw_sprite_h_flip(buffer, stand, s_x, s_y);
      }
    }
  }
  
  void Update(Level &level) {
    static bool skipspace = false;
    float gravity = -.4;
    v_y += gravity;
    y += v_y;
    
    float velx = 5.0f;
    if (key[KEY_RIGHT]) {
      if (ABS(anim_time - last_right) > 9 && key[KEY_SPACE] && skipspace) {
        v_x = velx*5;
        last_right = anim_time;
      }
      else {
        if (v_x > velx)
          v_x *= .97;
        if (v_x < velx || ABS(anim_time - last_right) > 3)
          v_x = velx;
      }
      state = RUN;
      last = 1;
    }
    if (key[KEY_LEFT]) {
      if (ABS(anim_time - last_left) > 9 && key[KEY_SPACE] && skipspace) {
        v_x = -velx*5;
        last_left = anim_time;
      }
      else {
        if (v_x < -velx)
          v_x *= .97;
        if (v_x > -velx || ABS(anim_time - last_left) > 3)
          v_x = -velx;
      }
      state = RUN;
      
      last = -1;
    }
    if (!key[KEY_LEFT] && !key[KEY_RIGHT]) {
      if (state != JUMP) 
        v_x = 0;
      state = STAND;
    }
    if (key[KEY_DOWN] && state != JUMP) {
      v_x = .99 * v_x;
      state = CROUCH;
    }
    
    int m_y = level.MaxY(int(x), int(x+stand->w));
    x+=v_x;
    int s_m_y = level.MaxY(int(x), int(x+stand->w));
    if (s_m_y - m_y > stand->h/3 && y < s_m_y)
      x -= v_x;
    
    if (x < 0) {
      if (state == RUN)
        state = STAND;
      x = 0;
      v_x = 0;
    }
    if (x + stand->w > SCREEN_W-1) {
      if (state == RUN)
        state = STAND;
      x = SCREEN_W - 1 - stand->w;
      v_x = 0;
    }
    
    m_y = level.MaxY(int(x), int(x+stand->w));
    if (m_y >= y) {
      v_y = 0;
      y = m_y;
      if (state == JUMP)
        state = STAND;
      
      if (key[KEY_UP]) {
        v_y = 15.0f;
        state = CROUCH;
      }
    }
    else {
      state = JUMP;
    }
    if (state == STAND && v_x != 0)
      state = RUN;
    
    skipspace = !key[KEY_SPACE];
  }
  
 
};

// My conscience DEMANDS I inform YOU THE READER that
// GLASS AND BOOT STRUCTS are just ASKING for TROUBLE because they DESTRUCT static resources when they have NO GUARANTEE some other user doesn't know about them
//...but hey, this is ludum dare.  Do what I say, kids, not what I do.

const int glass_c_x = 43,
              glass_c_y = 43,
              glass_radius = 40;
struct Glass {
  float x, y, v_x, v_y;
  float factor;
  
  static BITMAP *glass, *pattern;
  BITMAP *workspace;
  
  Glass(int x, int y) : x(x), y(y), v_x(0), v_y(0) {
    if (!glass)
      glass = LoadBitmap("mag.bmp");
    if (!pattern)
      pattern = LoadBitmap("mag_pattern.bmp");
    
    workspace = create_bitmap(glass->w, glass->h);
    factor = .01f + .05 * (rand()%10);
  }
  
  ~Glass() {
    destroy_bitmap(workspace);
    workspace = NULL;
  }
  
  static void Clean() {
    destroy_bitmap(glass);
    destroy_bitmap(pattern);
    glass = NULL;
    pattern = NULL;
  }
  
  void Update(Character &c, list<Particle> &parts) { // returns true if burning
    if (rand() % 30 == 0)
      factor *= 10.0/11.0;
    if (rand() % 30 == 0)
      factor *= 1.1;
    if (factor > 2.0)
      factor = 2.0;
    
    float dx = (c.x + c.stand->w/2) - x;
    float dy = (c.y + c.stand->h/2) - y;
    
    float dist = sqrt(dx * dx + dy * dy);
    if ( dist < float(glass_radius) / 2.0f ) {
      c.hit = true;
      for (int i = 0; i < (rand() % 3) + 3; i++) {
        parts.push_back(Particle(x + (rand()%(glass_radius)) - glass_radius/2, y + (rand()%(glass_radius)) - glass_radius/2, makecol(150 + rand()%100, 100 + rand()%50, 0), .01, .1, 1000));
        parts.push_back(Particle(x + (rand()%(glass_radius)) - glass_radius/2, y + (rand()%(glass_radius)) - glass_radius/2, makecol(150 + rand()%100, 100 + rand()%50, 0), .01, .1, 1000));
      }
    }
    if (dist > 2) {
      dx /= dist;
      dy /= dist;
      v_x += dx * factor;
      v_y += dy * factor;
    }
    x+=v_x;
    y+=v_y;
    v_y *= .95;
    v_x *= .95;
  }
  
  void Draw(BITMAP *buffer, int why) {
    int y_s = int(why) - int(y) - glass_c_y + HEIGHT;
    int x_s = int(x - glass_c_x);
    int y_c_s = int(y_s + glass_c_y);
    
    int radius = glass_radius;
    if (int(x) - radius < 0)
      radius = int(x);
    if (int(x) + radius >= SCREEN_W)
      radius = SCREEN_W-1-int(x);
    if (y_c_s - radius < 0)
      radius = y_c_s;
    if (y_c_s + radius >= SCREEN_H)
      radius = SCREEN_H-1-y_c_s;
    
    
    if (!buffer || !workspace || !glass || !pattern) {
      cout << "WHAT HAPPENED TO MY BMPS!?" << endl;
      exit(1);
    }
    if (radius > 0)
      stretch_blit(buffer, workspace,  int(x) - radius, y_c_s - radius, radius*2, radius*2, glass_c_x - radius*2, glass_c_y - radius*2, radius*4, radius*4);
    draw_sprite(workspace, glass, 0, 0);
    drawing_mode(DRAW_MODE_MASKED_PATTERN, pattern, 0, 0);
    rectfill(workspace, 0, 0, workspace->w, workspace->h, makecol(255,0,255));
    solid_mode();
    draw_sprite(buffer, workspace, x_s, y_s);
  }
};

BITMAP * Glass::glass = NULL;
BITMAP * Glass::pattern = NULL;

struct Boot {
  float x, y, v_y;
  bool done;
  
  static BITMAP *shadow, *boot;
  
  Boot(int x, Level &level) : x(x), v_y(0), done(false) {
    if (!shadow)
      shadow = LoadBitmap("shadow.bmp");
    if (!boot)
      boot = LoadBitmap("foot.bmp");
    y = level.MaxY(int(x), int(x + boot->w)) + SCREEN_H*2;
  }
  
  static void Clean() {
    delete shadow;
    delete boot;
    shadow = NULL;
    boot = NULL;
  }
  
  bool Hit(Character &c) {
    int height = c.stand->h - 20;
    if (c.state == CROUCH)
      height = 3*c.stand->h/5;
    return (c.x+c.stand->w/2 < x + boot->w && c.x + c.stand->w/2 > x && (y - boot->h - c.y) < height && (y - boot->h - c.y) > 0);
  }
  
  void Set(int ex, Level &level) {
    x = ex;
    v_y = 0;
    done = false;
    y = level.MaxY(int(x), int(x + boot->w)) + SCREEN_H*2;
  }
  
  void Update(Level &level) {
    if (done)
      return;
    
    v_y -= .2f;
    y += v_y;
    
    int m_y = level.MaxY( int(x), int(x + boot->w) );
    if (y - boot->h < m_y - 5) {
      level.Crush(int(x), int(x + boot->w), 20);
      done = true;
    }
  }
  
  void Draw(BITMAP *buffer, float why, Level &level) {
    if (done)
      return;
    
    draw_sprite(buffer, boot, int(x), int(why) - int(y) + HEIGHT);
    for (int ex = 0; ex < shadow->w; ex++) {
      masked_blit(shadow, buffer, ex, 0,  int(x) + ex, -level.heights[int(x)+ex] + int(why) + HEIGHT - 12, 1, shadow->h);
    }
  }
};

BITMAP * Boot::shadow = NULL;
BITMAP * Boot::boot = NULL;



struct Score {
  Score(string name, int score) : name(name), score(score) {}
  Score() : score(0) {}
  
  string name;
  int score;
};


#define MAX_SCORES 10

void HighScores(BITMAP *buffer, int score, FONT *f) {
  list<Score> scores;
  
  ifstream file("scores");
  string temp;
  int worst = -1;
  int count = 0;
  int rank = -1;
  char c;
  while (count < MAX_SCORES && getline(file, temp)) {
    if (temp == "")
      break;
    int s;
    file >> s;
    do {
      c = file.get();
    } while (c != '\n');
    if (rank < 0 && s < score)
      rank = count + 1;
    if (worst < 0 || s < worst)
      worst = s;
    scores.push_back(Score(temp, s));
    count++;
  }
  if (rank < 0)
    rank = count;
  if (rank == 0)
    rank = 1;
  file.close();

  if (count < MAX_SCORES-1 || score > worst) {
    char name[15];
    name[0] = '\0';
    int c = 0, t = 0;
    char ascii = 0;
    while (key[KEY_ENTER] || key[KEY_ESC]) {}
    clear_keybuf();
    while (!key[KEY_ENTER] || c == 1) {
      if (key[KEY_ESC] || quitnow)
        exit(0);
      
      clear_to_color(buffer, makecol(255,255,255));
      textprintf_centre(buffer, f, SCREEN_W/2, 50, makecol(0,0,0),  "High Score #%d / %d!  You survived %d seconds.", rank, MAX_SCORES, score);
      textprintf_centre(buffer, f, SCREEN_W/2, SCREEN_H/2, makecol(0,0,0),  "Enter your name: %s", name);
      blit(buffer, screen, 0, 0, 0, 0, SCREEN_W, SCREEN_H);
      
      while (!keypressed()) {
        if (quitnow)
          exit(0);
      }
      
      int k = readkey();
      
      t++;
      
      if (k >> 8 == KEY_BACKSPACE) {
        if (c > 0)
          c--;
        name[c] = '\0';
      }
      
      ascii = k & 0xff;
      if (c < 14 && ((ascii >= 'A' && ascii <= 'Z') || ascii == ' ' || (ascii >= 'a' && ascii <= 'z') || (ascii >= '0' && ascii <= '9'))) {
        name[c] = ascii;
        name[++c] = '\0';
      }
    }
    
    clear_to_color(buffer, makecol(255,255,255));
    textprintf_centre(buffer, f, SCREEN_W/2, 20, makecol(0,0,0),  "Hall of Awesome");
    
    ofstream out("scores");
    int count = 0;
    bool outyet = false;
    for (list<Score>::iterator i = scores.begin(); i != scores.end(); ++i) {
      if (!outyet && i->score < score) {
        if (count++ < MAX_SCORES)
          textprintf_centre(buffer, f, SCREEN_W/2, 55 + 35 * count, makecol(0,0,0),  "%s -- %d", name, score);
        out << name << "\n" << score << endl;
        outyet = true;
      }
      out << i->name << "\n" << i->score << endl;
      if (count++ < MAX_SCORES)
        textprintf_centre(buffer, f, SCREEN_W/2, 55 + 35 * count, makecol(0,0,0),  "%s -- %d", i->name.c_str(), i->score);
    }
    if (!outyet) {
      out << name << "\n" << score << endl;
      if (count++ < MAX_SCORES)
        textprintf_centre(buffer, f, SCREEN_W/2, 55 + 35 * count, makecol(0,0,0),  "%s -- %d", name, score);
      outyet = true;
    }
    out.close();
    
    textprintf_centre(buffer, f, SCREEN_W/2, 55 + 35 * (MAX_SCORES+2), makecol(0,0,0),  "Press [Enter] to continue");
    
    while (key[KEY_ENTER] || key[KEY_ESC]) {}
    clear_keybuf();
    while (!key[KEY_ENTER]) {
      if (key[KEY_ESC] || quitnow)
        exit(0);
      
      blit(buffer, screen, 0, 0, 0, 0, SCREEN_W, SCREEN_H);
    }
  }
}



int main() {
  
  BITMAP *buffer = init(800,600);
  if (!buffer)
    exit(1);
  
	// get a pretty font
  DATAFILE* fontdat = load_datafile("font.dat");
  if (!fontdat) {
    allegro_message("Failed to load font datafile");
    exit(1);
  }
  FONT* newfont = (FONT*)fontdat[0].dat;
  if (!newfont) {
    allegro_message("Failed to load font");
    exit(1);
  }
  text_mode(-1);
  
  BITMAP *splash = LoadBitmap("splash.bmp");
  while (key[KEY_ENTER]) {}
  clear_keybuf();
  while (!key[KEY_ENTER]) {
    if (key[KEY_ESC] || quitnow)
      exit(0);
    
    blit(splash, buffer, 0, 0, 0, 0, SCREEN_W, SCREEN_H);
    if (logic_time > 60)
      textprintf_centre(buffer, newfont, SCREEN_W/2, SCREEN_H-40, makecol(0,0,0),  "Press [Enter] to continue");
    blit(buffer, screen, 0, 0, 0, 0, SCREEN_W, SCREEN_H);
  }
  destroy_bitmap(splash);
  splash = NULL;
  
  while (!key[KEY_ESC] && !quitnow){
    srand(27); // it's only pseudorandom ...
    install_int_ex(anim_timer, BPS_TO_TIMER(30)); // start the anim timer again
    Level level;
    Character me;
    list<Glass*> glasses;
    glasses.push_back(new Glass(SCREEN_W/4,0));
    
    list<Boot> boots;
    boots.push_back(Boot(SCREEN_W/2, level));
    list<Particle> parts;
    
    int score = 0;
    logic_time = anim_time = 0;
    bool dead = false;
    
    while (!key[KEY_ESC] && !quitnow && !(dead==true && key[KEY_ENTER])) {
      while (logic_time > 0) {
        if (!dead) {
          if (boots.size()*boots.size() < unsigned(score / 700.0 + 1) && !(boots.size() > 5)) {
            boots.push_back(Boot(int(me.x), level));
          }
          if (glasses.size()*glasses.size() < unsigned(score / 500.0 + 1) && !(glasses.size() > 7)) {
            glasses.push_back(new Glass(SCREEN_W/2, int(me.y) + SCREEN_H*2));
          }
          
          me.Update(level);
          for (list<Boot>::iterator i = boots.begin(); i != boots.end(); ++i) {
            i->Update(level);
          }
          me.hit = false;
          for (list<Glass*>::iterator i = glasses.begin(); i != glasses.end(); ++i) {
            (*i)->Update(me, parts);
            //g2.Update(me, parts);
          }
          if (me.hit) {
            me.health--;
          }
          else {
            me.health++;
          }
          #define MAX_HEALTH 8
          if (me.health > MAX_HEALTH)
            me.health = MAX_HEALTH;
          else if (me.health < 0) {
            dead = true;
            me.Boom(parts, makecol(255,50,0));
            remove_int(anim_timer);
          }
          for (list<Boot>::iterator boot = boots.begin(); boot != boots.end(); ++boot) {
            if (boot->done) {
              boot->Set(int(me.x) + rand()%80 - 40, level);
            }
            else {
              if (boot->Hit(me)) {
                if (me.state == JUMP) {
                  me.v_y = boot->v_y*2 < me.v_y ? boot->v_y*2 : me.v_y;
                }
                else {
                  dead = true;
                  me.Boom(parts, makecol(0,0,0));
                  remove_int(anim_timer);
                }
              }
            }
          }
          score++;
        }
        for (list<Particle>::iterator i = parts.begin(); i != parts.end();) {
          i->Update();
          if (i->life <= 0)
            i = parts.erase(i);
          else
            ++i;
        }
        logic_time--;
      }
      
      vsync();
      level.Draw(buffer, me.y);
      if (!dead)
        me.Draw(buffer);
      for (list<Particle>::iterator i = parts.begin(); i != parts.end(); ++i) {
        i->Draw(buffer, int(me.y));
      }
      for (list<Boot>::iterator boot = boots.begin(); boot != boots.end(); ++boot) {
        boot->Draw(buffer, me.y, level);
      }
      for (list<Glass*>::iterator g = glasses.begin(); g != glasses.end(); ++g) {
        (*g)->Draw(buffer, int(me.y));
      }
      if (!dead)
        textprintf_centre(buffer, newfont, SCREEN_W/2, 20, makecol(0,0,0),  "Survived: %d seconds", score/60);
      else {
        textprintf_centre(buffer, newfont, SCREEN_W/2, SCREEN_H/2, makecol(255,0,0),  "You are dead.");
        textprintf_centre(buffer, newfont, SCREEN_W/2, SCREEN_H-40, makecol(0,0,0),  "Press [Enter] to continue");
      }
      blit(buffer, screen, 0, 0, 0, 0, SCREEN_W, SCREEN_H);
      clear_to_color(buffer, makecol(0,255,255));
    }
    
    for (list<Glass*>::iterator i = glasses.begin(); i != glasses.end(); ++i) {
      delete *i;
    }
    
    HighScores(buffer, score / 60, newfont);
  }
  
  Glass::Clean();
  Boot::Clean();
  
  return 0;
}END_OF_MAIN()

